﻿
#region Header

/* /////////////////////////////////////////////////////////////////////////////////////////////////
XpVRControllers for Virt-A-Mate by Fabio Den.

Written in Visual Studio
	
Version 1.0 2022-03-09
	Initial release.

License: CC BY-SA
	
///////////////////////////////////////////////////////////////////////////////////////////////// */

using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using UnityEngine.UI;
using UnityEngine.Events;
using UnityEngine.Animations;
using System.Linq;
using MVR.FileManagementSecure;
using SimpleJSON;

#endregion

public class VRControllersFree : MVRScript
{

	#region Variables

	private JSONStorableBool enabledChk;
	private float tolerance = 0.1f;
	private float tolerancer = 0.35f;
	private bool changing = false;
	private bool changing2 = false;
	private bool wasLB = false;
	private int sep = 5;

	private string leftAtom = "";
	private string rightAtom = "";
	private FreeControllerV3 leftObject = null;	 		// selected left controller
	private FreeControllerV3 rightObject = null;	 	// selected right controller
	private bool rotatingL = false;
	private bool rotatingR = true;

	private JSONStorableStringChooser atomPopup = null;
	private JSONStorableStringChooser jointPopup;

	private UIDynamicTextField myLabelL;
	private UIDynamicTextField myLabelR;

	private JSONStorableBool localCoordL;
	private JSONStorableBool localCoordR;

	private JSONStorableFloat moveSpeedL;
	private JSONStorableFloat moveSpeedR;
	private JSONStorableFloat rotSpeedL;
	private JSONStorableFloat rotSpeedR;

	private JSONStorableBool disableRoot;
	
	#endregion

	#region Events

	public override void Init()
	{
		SetupGUI();
	}

	public void OnEnable()
	{
		SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
	}

	public void OnDisable()
	{
		SuperController.singleton.onSceneLoadedHandlers -= OnSceneLoaded;
	}

	protected void OnSceneLoaded()
	{
		leftAtom = "";
		rightAtom = "";

		leftObject = null;
		rightObject = null;

		rotatingL = false;
		rotatingR = false;

		atomPopup.val = "";
		jointPopup.val = "";

		UpdateStatus();
	}

	public void OnDestroy()
	{
		if (enabledChk.val)
		{
			SuperController.singleton.ToggleDisableAllNavigation();
		}
	}

	#endregion

	#region GUI

	private void SetupGUI()
	{
		CreateSpacer(false).height = sep;
		CreateSpacer(true).height = sep;

		enabledChk = GUICreateCheckbox(this, "Enabled", "Enabled", false, false, SyncEnabled);
		disableRoot = GUICreateCheckbox(this, "disableRoot", "Disable person 'root' movement?", true, true, SyncRoot);

		CreateSpacer(false).height = sep;
		CreateSpacer(true).height = sep;

		atomPopup = GUICreatePopup(this, "atomPopup", "Select  Person", false, SyncTempPopupChoices, SyncTempPopup);
		jointPopup = GUICreatePopup(this, "jointPopup", "Select  Joint", true, SyncJointPopupChoices, SyncJointPopup);

		myLabelL = CreateTextField(new JSONStorableString("", ""), false);
		myLabelL.text = " ";
		myLabelL.height = 120;

		myLabelR = CreateTextField(new JSONStorableString("", ""), true);
		myLabelR.text = " ";
		myLabelR.height = 120;

		GUICreateButton(this, "Map to Left Controller", new Color(0.78F, 0.78F, 0.899F), true, false, OnMapToLeft);
		GUICreateButton(this, "Map to Right Controller", new Color(0.78F, 0.78F, 0.899F), true, true, OnMapToRight);

		GUICreateButton(this, "Clear", new Color(0.78F, 0.78F, 0.899F), true, false, OnClearLeft);
		GUICreateButton(this, "Clear", new Color(0.78F, 0.78F, 0.899F), true, true, OnClearRight);

		CreateSpacer(false).height = sep;
		CreateSpacer(true).height = sep;

		moveSpeedL = GUICreateSlider(this, "moveSpeedL", "Left Movement Speed", 35.0F, false, 0F, 100.0F);
		moveSpeedR = GUICreateSlider(this, "moveSpeedR", "Right Movement Speed", 35.0F, true, 0F, 100.0F);

		rotSpeedL = GUICreateSlider(this, "rotSpeedL", "Left Rotation Speed", 25.0F, false, 0F, 100.0F);
		rotSpeedR = GUICreateSlider(this, "rotSpeedR", "Right Rotation Speed", 25.0F, true, 0F, 100.0F);

		localCoordL = GUICreateCheckbox(this, "localCoordL", "Use Local Coordinates", false, false, SyncCoordL);
		localCoordR = GUICreateCheckbox(this, "localCoordR", "Use Local Coordinates", false, true, SyncCoordR);

		UpdateStatus();
	}

	// before opening the popup
	private void SyncTempPopupChoices()
	{
		List<string> atomChoices = getAtomChoices();
		atomPopup.choices = atomChoices;
	}

	// after choosing a string from the popup
	private void SyncTempPopup(string atomUID)
	{
		atomPopup.val = atomUID;
	}

	public List<string> getAtomChoices()
	{
		List<string> choices = new List<string>();

		foreach (Atom a in SuperController.singleton.GetAtoms())
		{
			if (a.category == "People")
				choices.Add(a.name);
		}

		return choices;
	}

	// before opening the popup
	private void SyncJointPopupChoices()
	{
		List<string> mychoices = new List<string>();

		mychoices.Add("headControl");
		mychoices.Add("chestControl");
		mychoices.Add("rHandControl");
		mychoices.Add("lHandontrol");
		mychoices.Add("hipControl");
		mychoices.Add("penisBaseControl");
		mychoices.Add("rFootControl");
		mychoices.Add("lFootControl");

		jointPopup.choices = mychoices;
	}

	// after choosing a string from the popup
	private void SyncJointPopup(string atomUID)
	{
		jointPopup.val = atomUID;
	}

	public static JSONStorableStringChooser GUICreatePopup(MVRScript dll, string name, string label, bool rightside, UIPopup.OnOpenPopup popup, JSONStorableStringChooser.SetStringCallback callback)
	{
		JSONStorableStringChooser gui = new JSONStorableStringChooser(name, null, "", "", callback);
		dll.RegisterStringChooser(gui);
		UIDynamicPopup dp = dll.CreateFilterablePopup(gui, rightside);
		dp.popupPanelHeight = 600.0F;
		dp.label = label;
		dp.popup.onOpenPopupHandlers = popup;

		return gui;
	}

	private void SetFontSize(UIDynamicButton btn, int fontsize)
	{
		Text text_1 = btn.button.GetComponentInChildren<UnityEngine.UI.Text>();
		if (text_1) { text_1.fontSize = fontsize; }
	}

	public static UIDynamicButton GUICreateButton(MVRScript dll, string label, Color _color, bool useColor, bool rightside, UnityAction callback)
	{
		UIDynamicButton gui = dll.CreateButton(label, rightside);
		if (useColor) { gui.buttonColor = _color; }
		gui.button.onClick.AddListener(callback);

		return gui;
	}

	public static JSONStorableFloat GUICreateSlider(MVRScript dll, string name, string label, float startvalue, bool rightside, float minvalue, float maxvalue)
	{
		JSONStorableFloat gui = new JSONStorableFloat(name, startvalue, minvalue, maxvalue);
		var Slider = dll.CreateSlider(gui, rightside);
		Slider.label = label;
		dll.RegisterFloat(gui);

		return gui;
	}

	public static JSONStorableBool GUICreateCheckbox(MVRScript dll, string name, string label, bool startvalue, bool rightside, JSONStorableBool.SetJSONBoolCallback callback)
	{
		JSONStorableBool gui = new JSONStorableBool(name, startvalue, callback);
		var Toggle = dll.CreateToggle(gui, rightside);
		Toggle.label = label;

		return gui;
	}

	#endregion

	#region	Status

	private void UpdateStatus()
	{
		string leftCont = "";
		string rightCont = "";

		if (leftObject == null) 
		{ 
			leftAtom = "Not assigned"; 
		}
		else
		{
			leftCont = leftObject.name;
		}

		if (rightObject == null) 
		{ 
			rightAtom = "Not assigned"; 
		} 
		else
		{
			rightCont = rightObject.name;
		}

		bool isrotL = isRotating(false);
		string rotL = "\nStatus: <b>Moving</b> / Rotating";
		if (isrotL) { rotL = "\nStatus: Moving / <b>Rotating</b>"; }

		myLabelL.text = "Target atom: <b>" + leftAtom + "</b>\nController: " + leftCont + rotL;

		bool isrotR = isRotating(true);
		string rotR = "\nStatus: <b>Moving</b> / Rotating";
		if (isrotR) { rotR = "\nStatus: Moving / <b>Rotating</b>"; }

		myLabelR.text = "Target atom: <b>" + rightAtom + "</b>\nController: " + rightCont + rotR;
	}

	#endregion

	#region	Mapping

	private void OnMapToLeft()
	{
		Atom cur = SuperController.singleton.GetAtomByUid(atomPopup.val);
		FreeControllerV3 controller = null;

		if (cur != null)
		{
			controller = cur.GetStorableByID(jointPopup.val) as FreeControllerV3;
		}

		if (controller == null)
		{
			cur = SuperController.singleton.GetSelectedAtom();
			if (cur == null) { return; }

			controller = SuperController.singleton.GetSelectedController();
			if (controller == null) { return; }	
		}

		leftAtom = cur.name;
		leftObject = controller;
		rotatingL = false;

		UpdateStatus();
	}

	private void OnMapToRight()
	{
		Atom cur = SuperController.singleton.GetAtomByUid(atomPopup.val);
		FreeControllerV3 controller = null;

		if (cur != null)
		{
			controller = cur.GetStorableByID(jointPopup.val) as FreeControllerV3;
		}

		if (controller == null)
		{
			cur = SuperController.singleton.GetSelectedAtom();
			if (cur == null) { return; }

			controller = SuperController.singleton.GetSelectedController();
			if (controller == null) { return; }
		}

		rightAtom = cur.name;
		rightObject = controller;
		rotatingR = false;

		UpdateStatus();
	}

	private void OnClearLeft()
	{
		leftAtom = "";
		leftObject = null;
		rotatingL = false;

		atomPopup.val = "";
		jointPopup.val = "";

		UpdateStatus();
	}

	private void OnClearRight()
	{
		rightAtom = "";
		rightObject = null;
		rotatingR = false;

		atomPopup.val = "";
		jointPopup.val = "";

		UpdateStatus();
	}

	#endregion

	#region	States

	private float GetSpeed(bool rightSide)
	{
		if (rightSide) { return moveSpeedR.val; }
		return moveSpeedL.val;
	}

	private bool usingLocal(bool rightSide)
	{
		if (rightSide) { return localCoordR.val; }
		return localCoordL.val;
	}

	private FreeControllerV3 GetObject(bool rightSide)
	{
		FreeControllerV3 controller = SuperController.singleton.GetSelectedController();

		if ((leftObject != null) || (rightObject != null))
		{
			// controller = null;   // confusing either way
		}

		if (rightSide)
		{
			if (rightObject != null) {	controller = rightObject; }
		}
		else
		{
			if (leftObject != null) { controller = leftObject; }
		}

		if (disableRoot.val)
		{
			if (controller.name.Equals("control")) 
			{
				Atom cur = SuperController.singleton.GetSelectedAtom();
				if (cur == null) { return null; }

				if (cur.category == "People")
					controller = null;
			}
		}

		return controller;
	}

	private float GetRotSpeed(bool rightSide)
	{
		if (rightSide) { return rotSpeedR.val; }
		return rotSpeedL.val;
	}

	private bool isRotating(bool rightside)
	{
		if (rightside) { return rotatingR; }
		return rotatingL;
	}

	private bool isPossessing(bool rightside)
	{
		return false;
	}

	#endregion

	#region	Movement

	public void ControllerMove(int dir, float up, bool rightSide)
	{
		FreeControllerV3 controller = GetObject(rightSide);
		if (controller == null) { return; }

		var direction = Vector3.right;

		switch (dir)
		{
			case 1:
				direction = Vector3.right;
				break;
			case 2:
				direction = Vector3.up;
				break;
			case 3:
				direction = Vector3.forward;
				break;
		}

		float speed = GetSpeed(rightSide) / 100F;
		bool local = usingLocal(rightSide);
		
		if (local)
			controller.transform.Translate(direction * Time.deltaTime * up * speed, Space.Self);
		else
			controller.transform.Translate(direction * Time.deltaTime * up * speed, Camera.main.transform);
	}

	#endregion

	#region	Rotation

	public void ControllerRotate(int dir, float up, bool rightSide)
	{
		float speed = GetRotSpeed(rightSide) / 100F;
		bool local = usingLocal(rightSide);

		var rotate = 360 * Time.deltaTime * up * speed;

		FreeControllerV3 target = GetObject(rightSide);
		if (target == null) { return; }

		if (local)
		{
			if (dir == 1) { target.transform.Rotate(rotate, 0, 0, Space.Self); } // yes
			else if (dir == 2) { target.transform.Rotate(0, rotate, 0, Space.Self); }// no
			else if (dir == 3) { target.transform.Rotate(0, 0, rotate, Space.Self); } // roll	
		}
		else
		{
			if (dir == 1) { target.transform.RotateAround(target.transform.position, Camera.main.transform.right, -rotate); } // yes
			else if (dir == 2) { target.transform.RotateAround(target.transform.position, Camera.main.transform.up, rotate); } // no
			else if (dir == 3) { target.transform.RotateAround(target.transform.position, Camera.main.transform.forward, rotate); } // roll
		}
	}

	#endregion

	#region	Sticks

	private void ProcessLeftStick(bool a, Vector4 axes)
	{
		if (a == false)
		{
			// LEFT STICK HORIZONTAL
			if (isRotating(false) == false)
			{
				if (Math.Abs(axes.x) > tolerance)
				{
					ControllerMove(1, axes.x, false);
				}
			}
			else
			{
				if (Math.Abs(axes.x) > tolerancer)
				{
					ControllerRotate(2, -axes.x, false);
				}
			}

			// LEFT STICK VERTICAL
			if (isRotating(false) == false)
			{
				if (Math.Abs(axes.y) > tolerance)
				{
					ControllerMove(2, axes.y, false);
				}
			}
			else
			{
				if (Math.Abs(axes.y) > tolerancer)
				{
					ControllerRotate(1, -axes.y, false);
				}
			}
		}
		else
		{
			if (isRotating(false) == false)
			{
				// LEFT STICK VERTICAL
				if (Math.Abs(axes.y) > tolerance)
				{
					ControllerMove(3, axes.y, false);
				}
			}
			else
			{
				// LEFT STICK HORIZONTAL
				if (Math.Abs(axes.x) > tolerancer)
				{
					ControllerRotate(3, -axes.x, false);
				}
			}
		}
	}

	private void ProcessRightStick(bool b, Vector4 axes)
	{
		if (b == false)
		{
			// RIGHT STICK HORIZONTAL
			if (isRotating(true) == false)
			{
				if (Math.Abs(axes.z) > tolerance)
				{
					ControllerMove(1, axes.z, true);
				}
			}
			else
			{
				if (Math.Abs(axes.z) > tolerancer)
				{
					ControllerRotate(2, -axes.z, true);
				}
			}


			// RIGHT STICK VERTICAL
			if (isRotating(true) == false)
			{
				if (Math.Abs(axes.w) > tolerance)
				{
					ControllerMove(2, axes.w, true);
				}
			}
			else
			{
				if (Math.Abs(axes.w) > tolerancer)
				{
					ControllerRotate(1, -axes.w, true);
				}
			}
		}
		else
		{
			if (isRotating(true) == false)
			{
				// RIGHT STICK VERTICAL
				if (Math.Abs(axes.w) > tolerance)
				{
					ControllerMove(3, axes.w, true);
				}
			}
			else
			{
				// RIGHT STICK HORIZONTAL
				if (Math.Abs(axes.z) > tolerancer)
				{
					ControllerRotate(3, -axes.z, true);
				}
			}
		}
	}

	#endregion

	#region	Flags

	private void SyncRoot(JSONStorableBool flag)
	{
		disableRoot.val = flag.val;
	}

	private void SyncCoordL(JSONStorableBool flag)
	{
		localCoordL.val = flag.val;
	}

	private void SyncCoordR(JSONStorableBool flag)
	{
		localCoordR.val = flag.val;
	}

	private void SyncEnabled(JSONStorableBool flag)
	{
		enabledChk.val = flag.val;

		SuperController.singleton.ToggleDisableAllNavigation();
	}

	#endregion

	#region Update

	void FixedUpdate()
	{
		DoStep();
	}


	private void DoStep()
	{
		bool LB = SuperController.singleton.GetLeftHoldGrab();
		bool RB = SuperController.singleton.GetRightHoldGrab();
		bool LT = SuperController.singleton.GetLeftGrabVal() > tolerance;
		bool RT = SuperController.singleton.GetRightGrabVal() > tolerance;

		// Enable - disable the plugin
		if ((LB && LT) || (RB && RT))
		{
			if (changing == false)
			{
				changing = true;
				enabledChk.val = !enabledChk.val;
				changing2 = false;
			}

			return;
		}

		// Preparing a future toggle move - rotate
		if (LB || RB)
		{
			if (changing2 == false)
			{
				changing2 = true;

				if (LB)
				{
					wasLB = true;
				}

				if (RB)
				{
					wasLB = false;
				}
			}

			return;
		}

		// toggle move - rotate
		if ((changing2 == true) && (changing == false) )
		{
			if (wasLB)
			{
				if (isPossessing(false) == false)
				{
					rotatingL = !rotatingL;
					UpdateStatus();
				}
			}
			else
			{
				if (isPossessing(true) == false)
				{
					rotatingR = !rotatingR;
					UpdateStatus();
				}
			}
		}

		// if nothing else was pressed proceed with the sticks
		if ((changing2 == false) && (changing == false))
		{
			if (enabledChk.val)
			{
				Vector4 axes = SuperController.singleton.GetFreeNavigateVector(SuperController.singleton.freeMoveAction);

				if (isPossessing(false) == false)
				{
					ProcessLeftStick(LT, axes);
				}

				if (isPossessing(true) == false)
				{
					ProcessRightStick(RT, axes);
				}
			}
		}

		changing2 = false;
		changing = false;
	}

	#endregion

} // End class

